/****************************************************************************/
/* ZEPTOOS:zepto-info */
/*     This file is part of ZeptoOS: The Small Linux for Big Computers.
 *     See www.mcs.anl.gov/zeptoos for more information.
 */
/* ZEPTOOS:zepto-info */
/* */
/* ZEPTOOS:zepto-fillin */
/*     $Id: pread.c,v 1.1.1.1 2006/01/26 21:06:58 kazutomo Exp $
 *     ZeptoOS_Version: DEVELOPMENT-PRERELEASE
 *     ZeptoOS_Heredity: FOSS_ORIG
 *     ZeptoOS_License: GPL
 */
/* ZEPTOOS:zepto-fillin */
/* */
/* ZEPTOOS:zepto-gpl */
/*      Copyright: Argonne National Laboratory, Department of Energy,
 *                 and University of Chicago.  2004, 2005
 *      ZeptoOS License: GPL
 * 
 *      This software is free.  See the file COPYING
 *      for complete details on your rights to copy, modify, and use this
 *      software.
 */
/* ZEPTOOS:zepto-gpl */
/****************************************************************************/

/**
 * This benchmark program is designed to measure the performance of
 * random pread(lseek+read) from IBM BlueGene Compute node.
 *
 * Originaly written by Kazutomo Yoshii <kazutomo@mcs.anl.gov>
 * 
 * [compile]
 * $ mpicc -Wall     pread.c   -o pread
 *
 * [run]
 * $ dd if=/dev/urandom of=mytest.dat bs=1M count=1024
 * $ cqsub -t 30 -n 128 ./pread -s 1048576 -f mytest.dat -d /pvfs/kazutomo
 * 
 * -s pread buffer size
 * -f testfile
 * -d directory testfile is located at
 *
 * other options
 * -n number of iterations for pread
 * -L add extra load by stderr output
 * -e use lseek()+read() instead of pread()
 *

 * [Version History]
 *
 * * ??? ??? ?? 2005 - Kazutomo Yoshii <kazutomo@mcs.anl.gov>\n
 *   v0.1 - basic capability. not in cvs 
 *
 * * Mon Jan 16 2006 - Kazutomo Yoshii <kazutomo@mcs.anl.gov>\n
 *   v0.2 - use bginfo. switched to Allreduce for add-up 
 *          
 *
 * Future plan: this program is going to be merged to bgiobench
 * program.  it will be available from my page or the ZeptoOS project
 * page.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <time.h>
#include <unistd.h>
#include <string.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/time.h>

#include "mpi.h"
#include "bginfo.h"

#define PREAD_VERSION   "0.2"
/*
	NOTE: previous version 0.1 didn't have PREAD_VERSION 
*/

const char*  fmtsb(const char* fmt,...)
{
  static char buf[128];
  va_list ap;
  va_start(ap,fmt);
  vsnprintf(buf,sizeof(buf),fmt,ap);
  return buf;
}

#define die(msg)   {printf("Fatal error: %s(%d) : %s\n",__FILE__,__LINE__,msg); exit(1);}



#ifdef REALLY_DO_NOT_HAVE_PREAD
size_t _pread(int fd, void *buf, size_t count, off_t offset)
{
  int rc;
  rc = lseek( fd, offset, SEEK_SET);
  if( rc < 0 ) {
		return rc;
  }
  rc = read( fd, buf, count );
	if( rc != count ) {
		perror("read");
		return rc;
  }
  return rc;
}
#endif


#define DEFAULT_DATA_FN   "test.dat"
#define DEFAULT_NITERS    (100)
#define DEFAULT_BLOCKSIZE (16*1024*1024)  /* = 16MBytes  */
/*
	On a BG/L system with 32 CN / 1 ION ratio and each ION has 512MB,
	maximum buffer size on ION for each CN is less than 16MB.
	Per default, 4MB is assigned by CIO for each CN.
	So 16MB pread from all CNs, of course, ran out  ION buffer.
 */

/**
 * just return the file size of a file
 */
static int get_file_size(const char* fn)
{
	int  fd,rc;
  struct stat st;
	
  fd = open( fn, O_RDONLY );
  if( fd < 0 ) {
    perror("open");
		return 0;
  }
  rc = fstat( fd, &st );
  if(rc<0) {
    perror( "fstat" );
		return 0;
  }
	return st.st_size;
}

static double get_current_time()
{
	struct timeval tv;
	gettimeofday(&tv, NULL);
	return (double)tv.tv_sec + (double)tv.tv_usec/1000/1000;
}



static int adding_stderr_load;

static double pread_bench(const char* datafn, int niters, int blocksize, int use_pread)
{
  int  fd, i;
  ssize_t pos, rc;
  ssize_t nblocks;
  char* buf;
  double  t1,t2;

  buf = (char*)malloc( blocksize );
  if( ! buf ) {
    die("Unable to allocate memory");
  }

	nblocks = get_file_size(datafn) / blocksize;

  fd = open( datafn, O_RDONLY );
  if( fd < 0 ) {
    die(fmtsb("Unable to open %s",datafn ));
  }

  t1 = get_current_time();
  for(i=0; i<niters; i++ ) {
    pos = rand() % nblocks;

    /** 
		 * use stderr message output as an extra I/O workload.  stdout/err
		 * message are transfered via ciod on ION to the service node. they
		 * are eventually written down to filesystem(NFS) there 
		 */
    if( adding_stderr_load )      fprintf(stderr,"pos=%d\n", pos);

    if( use_pread) {
#ifdef REALLY_DO_NOT_HAVE_PREAD
      rc = _pread( fd, buf, blocksize, pos*blocksize );
#else
      rc = pread( fd, buf, blocksize, pos*blocksize );
#endif
      if( rc != blocksize ) {
				die("Failed to pread");
      }
      if( rc < 0 ) {
        perror( "pread" );
        goto out;
      }
    } else {
      rc = lseek( fd, pos*blocksize, SEEK_SET);
      if( rc < 0 ) {
        perror("lseek");
        goto out;
      }
      rc = read( fd, buf, blocksize );
      if( rc < -1 ) {
        perror("read");
        goto out;
      }
    }
  }
  t2 = get_current_time();

 out:
  close(fd);

  free(buf);

  return t2-t1;
}




static void usage(const char* progname)
{
  puts("");
  printf("Usage: %s [options]\n",progname);
  puts("");
  puts("[options]");
  puts("-h : shows this help");
  puts("-e : emulation mode. lseek()+read()");
  puts("-n : the number of iterations");
  puts("-s : blocksize in byte");
  puts("-f : filename");
  puts("-L : add extra IO workload by generating lots of stderr output");
  puts("");
}


/**
 * calculate the aggregated throughput and return it
 * unit is MB/sec
 */ 

static double calc_throughput(double size, double blocksize, double niters, double t) 
{
  double tmp;
  tmp = size* blocksize*niters; /* total amount of I/O from nodes */
  tmp /= (double)1024*1024; /* in MB */
  return  tmp/t;  /* MB/sec */
}

int main(int argc,char* argv[])
{
  int    c;
  int    niters = DEFAULT_NITERS;
  int    blocksize = DEFAULT_BLOCKSIZE;
  int    use_pread=1;   /** 0 means emulation with lseek() + read() */
  char*  data_fn = DEFAULT_DATA_FN;
  char   cwdstr[128];
  double timeinsec;
	double max_pread_time, min_pread_time, sum_pread_time;
  int    rank,size;


  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &size);
  srand( (unsigned)time(NULL) - rank );

  bginfo_init();  /* does collective operation, so all procs must call this */

  while(1) {
    c = getopt( argc, argv, "Lhed:n:s:f:" );
    if( c == -1 ) break;

    switch(c) {
    case 'h': 
      usage(argv[0]);
      exit(1);
    case 'd':
      if( access(optarg, R_OK|W_OK|X_OK) == 0 ) {
        chdir( optarg );
      }
      break;
    case 'L':
      adding_stderr_load=1;
      break;
    case 'e':
      use_pread=0;
      break;
    case 'n':
      niters = atoi(optarg);
      break;
    case 'f':
      data_fn = optarg;
      break;
    case 's':
      blocksize = atoi(optarg);
      break;
      printf ("?? getopt returned character code 0%o ??\n", c);
    }
  }
  
  if( !data_fn ) { 
    printf("Please speficy a valid file name.\n");
    exit(2);
  }

  if( rank==0 ) {
		puts("#");
		printf("# [PREAD benchmark - version %s]\n", PREAD_VERSION);
		puts("#");
    getcwd(cwdstr,sizeof(cwdstr));
    printf("# DIR=%s\n",cwdstr);
		printf("# FILE=\"%s\"\n", data_fn);
    printf("# SIZE=%d\n",size);
		printf("# DATASIZE=%d bytes\n", get_file_size(data_fn));
    printf("# BUFFERSIZE_PREAD=%d bytes\n", blocksize);
    printf("# NITERS=%d times\n", niters);
    printf("# BUILDTIME=\"%s %s\"\n",__DATE__, __TIME__); 
    if( use_pread==0 ) {
      puts("# PREAD=LSEEK+READ\n");
    }

    if( adding_stderr_load ) 
      printf("# ADDING_STDERR_LOAD=1\n"); 
    fflush(stdout);
  }

  MPI_Barrier(MPI_COMM_WORLD);

  timeinsec = pread_bench(data_fn, niters, blocksize, use_pread);

	/* add up */
	MPI_Allreduce(&timeinsec, &max_pread_time, 1, MPI_DOUBLE, MPI_MAX,
								MPI_COMM_WORLD);
	MPI_Allreduce(&timeinsec, &min_pread_time, 1, MPI_DOUBLE, MPI_MIN,
								MPI_COMM_WORLD);
	MPI_Allreduce(&timeinsec, &sum_pread_time, 1, MPI_DOUBLE, MPI_SUM,
								MPI_COMM_WORLD);

	if( rank==0 ) {
		puts(" ");
		printf("%d   %lf %lf %lf # size average(MB/s) min(MB/s) max(MB/s)\n", 
					 size, 
					 calc_throughput(size, blocksize, niters, sum_pread_time/(double)size),
					 calc_throughput(size, blocksize, niters, max_pread_time),
					 calc_throughput(size, blocksize, niters, min_pread_time)  );
		puts(" ");
    fflush(stdout);
	}

	bginfo_print_info();  // shows BG/L software/hardware information
	// bginfo_print_info_full(); // for detailed information
	
	MPI_Finalize();

  return 0;
}

/*
 * Local variables:
 *  c-indent-level: 2
 *  c-basic-offset: 2
 *  tab-width: 2
 *
 * vim: ts=2
 * End:
 */
